#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 20
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M, ATL_CINT N, const TYPE *A, ATL_CINT lda,
               const TYPE *X, TYPE *Y)
/*
 * 2x20 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N20=(N/20)*20, M2=(M/2)*2, lda20=lda*20;
   ATL_INT j;

   for (j=N20; j; j -= 20, A += lda20, Y += 20)
   {
      const float *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda,
                  *A5=A4+lda, *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda,
                  *A10=A9+lda, *A11=A10+lda, *A12=A11+lda, *A13=A12+lda,
                  *A14=A13+lda, *A15=A14+lda, *A16=A15+lda, *A17=A16+lda,
                  *A18=A17+lda, *A19=A18+lda;
      register float y0=ATL_rzero, y1=ATL_rzero, y2=ATL_rzero, y3=ATL_rzero,
                     y4=ATL_rzero, y5=ATL_rzero, y6=ATL_rzero, y7=ATL_rzero,
                     y8=ATL_rzero, y9=ATL_rzero, y10=ATL_rzero, y11=ATL_rzero,
                     y12=ATL_rzero, y13=ATL_rzero, y14=ATL_rzero, y15=ATL_rzero,
                     y16=ATL_rzero, y17=ATL_rzero, y18=ATL_rzero,
                     y19=ATL_rzero;
      ATL_INT i;
      prefY(Y+20+20-1);
      for (i=0; i < M2; i += 2)
      {
         const register float x0=X[i+0], x1=X[i+1];

         prefX(X);
         y0 += A0[i+0] * x0;
         prefA(A0);
         y1 += A1[i+0] * x0;
         prefA(A1);
         y2 += A2[i+0] * x0;
         prefA(A2);
         y3 += A3[i+0] * x0;
         prefA(A3);
         y4 += A4[i+0] * x0;
         prefA(A4);
         y5 += A5[i+0] * x0;
         prefA(A5);
         y6 += A6[i+0] * x0;
         prefA(A6);
         y7 += A7[i+0] * x0;
         prefA(A7);
         y8 += A8[i+0] * x0;
         prefA(A8);
         y9 += A9[i+0] * x0;
         prefA(A9);
         y10 += A10[i+0] * x0;
         prefA(A10);
         y11 += A11[i+0] * x0;
         prefA(A11);
         y12 += A12[i+0] * x0;
         prefA(A12);
         y13 += A13[i+0] * x0;
         prefA(A13);
         y14 += A14[i+0] * x0;
         prefA(A14);
         y15 += A15[i+0] * x0;
         prefA(A15);
         y16 += A16[i+0] * x0;
         prefA(A16);
         y17 += A17[i+0] * x0;
         prefA(A17);
         y18 += A18[i+0] * x0;
         prefA(A18);
         y19 += A19[i+0] * x0;
         prefA(A19);
         y0 += A0[i+1] * x1;
         y1 += A1[i+1] * x1;
         y2 += A2[i+1] * x1;
         y3 += A3[i+1] * x1;
         y4 += A4[i+1] * x1;
         y5 += A5[i+1] * x1;
         y6 += A6[i+1] * x1;
         y7 += A7[i+1] * x1;
         y8 += A8[i+1] * x1;
         y9 += A9[i+1] * x1;
         y10 += A10[i+1] * x1;
         y11 += A11[i+1] * x1;
         y12 += A12[i+1] * x1;
         y13 += A13[i+1] * x1;
         y14 += A14[i+1] * x1;
         y15 += A15[i+1] * x1;
         y16 += A16[i+1] * x1;
         y17 += A17[i+1] * x1;
         y18 += A18[i+1] * x1;
         y19 += A19[i+1] * x1;
      }
      if (M != M2)
      {
         const register float x0 = X[i];
         y0 += A0[i] * x0;
         y1 += A1[i] * x0;
         y2 += A2[i] * x0;
         y3 += A3[i] * x0;
         y4 += A4[i] * x0;
         y5 += A5[i] * x0;
         y6 += A6[i] * x0;
         y7 += A7[i] * x0;
         y8 += A8[i] * x0;
         y9 += A9[i] * x0;
         y10 += A10[i] * x0;
         y11 += A11[i] * x0;
         y12 += A12[i] * x0;
         y13 += A13[i] * x0;
         y14 += A14[i] * x0;
         y15 += A15[i] * x0;
         y16 += A16[i] * x0;
         y17 += A17[i] * x0;
         y18 += A18[i] * x0;
         y19 += A19[i] * x0;
      }
      #ifdef BETA0
         Y[0] = y0;
         Y[1] = y1;
         Y[2] = y2;
         Y[3] = y3;
         Y[4] = y4;
         Y[5] = y5;
         Y[6] = y6;
         Y[7] = y7;
         Y[8] = y8;
         Y[9] = y9;
         Y[10] = y10;
         Y[11] = y11;
         Y[12] = y12;
         Y[13] = y13;
         Y[14] = y14;
         Y[15] = y15;
         Y[16] = y16;
         Y[17] = y17;
         Y[18] = y18;
         Y[19] = y19;
      #else
         Y[0] += y0;
         Y[1] += y1;
         Y[2] += y2;
         Y[3] += y3;
         Y[4] += y4;
         Y[5] += y5;
         Y[6] += y6;
         Y[7] += y7;
         Y[8] += y8;
         Y[9] += y9;
         Y[10] += y10;
         Y[11] += y11;
         Y[12] += y12;
         Y[13] += y13;
         Y[14] += y14;
         Y[15] += y15;
         Y[16] += y16;
         Y[17] += y17;
         Y[18] += y18;
         Y[19] += y19;
      #endif
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N20; j; j--, A += lda, Y++)
   {
      const float *A0=A;
      register float y0=ATL_rzero;
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M2; i += 2)
      {
         const register float x0=X[i+0], x1=X[i+1];

         prefX(X);
         y0 += A0[i+0] * x0;
         prefA(A0);
         y0 += A0[i+1] * x1;
      }
      if (M != M2)
      {
         const register float x0 = X[i];
         y0 += A0[i] * x0;
      }
      #ifdef BETA0
         Y[0] = y0;
      #else
         Y[0] += y0;
      #endif
   }
}
