#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 12
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGER2K(ATL_CINT M, ATL_CINT N, const double *X, const double *Y, 
                const double *W, const double *Z, double *A, ATL_CINT lda)
/*
 * 24x12 unrolled r2_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N12=(N/12)*12, M24=(M/24)*24, lda12=lda*12;
   ATL_INT j;

   for (j=N12; j; j -= 12, A += lda12, Y += 12)
   {
      double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
             *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda, *A10=A9+lda,
             *A11=A10+lda;
      const register double y0=Y[0], z0=Z[0], y1=Y[1], z1=Z[1], y2=Y[2],
                            z2=Z[2], y3=Y[3], z3=Z[3], y4=Y[4], z4=Z[4],
                            y5=Y[5], z5=Z[5], y6=Y[6], z6=Z[6], y7=Y[7],
                            z7=Z[7], y8=Y[8], z8=Z[8], y9=Y[9], z9=Z[9],
                            y10=Y[10], z10=Z[10], y11=Y[11], z11=Z[11];
      ATL_INT i;
      prefY(Y+12+12-1);
      prefY(Z+12+12-1);
      for (i=0; i < M24; i += 24)
      {
         register double a0, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11;
         const double x0=X[i+0], w0=W[i+0], x1=X[i+1], w1=W[i+1], x2=X[i+2],
                      w2=W[i+2], x3=X[i+3], w3=W[i+3], x4=X[i+4], w4=W[i+4],
                      x5=X[i+5], w5=W[i+5], x6=X[i+6], w6=W[i+6], x7=X[i+7],
                      w7=W[i+7], x8=X[i+8], w8=W[i+8], x9=X[i+9], w9=W[i+9],
                      x10=X[i+10], w10=W[i+10], x11=X[i+11], w11=W[i+11],
                      x12=X[i+12], w12=W[i+12], x13=X[i+13], w13=W[i+13],
                      x14=X[i+14], w14=W[i+14], x15=X[i+15], w15=W[i+15],
                      x16=X[i+16], w16=W[i+16], x17=X[i+17], w17=W[i+17],
                      x18=X[i+18], w18=W[i+18], x19=X[i+19], w19=W[i+19],
                      x20=X[i+20], w20=W[i+20], x21=X[i+21], w21=W[i+21],
                      x22=X[i+22], w22=W[i+22], x23=X[i+23], w23=W[i+23];

         prefX(X);
         a0 = A0[i+0];
         a0 += x0 * y0;
         a1 = A1[i+0];
         a1 += x0 * y1;
         a2 = A2[i+0];
         a2 += x0 * y2;
         a3 = A3[i+0];
         a3 += x0 * y3;
         a4 = A4[i+0];
         a4 += x0 * y4;
         a5 = A5[i+0];
         a5 += x0 * y5;
         a6 = A6[i+0];
         a6 += x0 * y6;
         a7 = A7[i+0];
         a7 += x0 * y7;
         a8 = A8[i+0];
         a8 += x0 * y8;
         a9 = A9[i+0];
         a9 += x0 * y9;
         a10 = A10[i+0];
         a10 += x0 * y10;
         a11 = A11[i+0];
         a11 += x0 * y11;
         prefX(W);
         a0 += w0 * z0;
         prefA(A0);
         A0[i+0] = a0;
         a1 += w0 * z1;
         prefA(A1);
         A1[i+0] = a1;
         a2 += w0 * z2;
         prefA(A2);
         A2[i+0] = a2;
         a3 += w0 * z3;
         prefA(A3);
         A3[i+0] = a3;
         a4 += w0 * z4;
         prefA(A4);
         A4[i+0] = a4;
         a5 += w0 * z5;
         prefA(A5);
         A5[i+0] = a5;
         a6 += w0 * z6;
         prefA(A6);
         A6[i+0] = a6;
         a7 += w0 * z7;
         prefA(A7);
         A7[i+0] = a7;
         a8 += w0 * z8;
         prefA(A8);
         A8[i+0] = a8;
         a9 += w0 * z9;
         prefA(A9);
         A9[i+0] = a9;
         a10 += w0 * z10;
         prefA(A10);
         A10[i+0] = a10;
         a11 += w0 * z11;
         prefA(A11);
         A11[i+0] = a11;
         a0 = A0[i+1];
         a0 += x1 * y0;
         a1 = A1[i+1];
         a1 += x1 * y1;
         a2 = A2[i+1];
         a2 += x1 * y2;
         a3 = A3[i+1];
         a3 += x1 * y3;
         a4 = A4[i+1];
         a4 += x1 * y4;
         a5 = A5[i+1];
         a5 += x1 * y5;
         a6 = A6[i+1];
         a6 += x1 * y6;
         a7 = A7[i+1];
         a7 += x1 * y7;
         a8 = A8[i+1];
         a8 += x1 * y8;
         a9 = A9[i+1];
         a9 += x1 * y9;
         a10 = A10[i+1];
         a10 += x1 * y10;
         a11 = A11[i+1];
         a11 += x1 * y11;
         a0 += w1 * z0;
         A0[i+1] = a0;
         a1 += w1 * z1;
         A1[i+1] = a1;
         a2 += w1 * z2;
         A2[i+1] = a2;
         a3 += w1 * z3;
         A3[i+1] = a3;
         a4 += w1 * z4;
         A4[i+1] = a4;
         a5 += w1 * z5;
         A5[i+1] = a5;
         a6 += w1 * z6;
         A6[i+1] = a6;
         a7 += w1 * z7;
         A7[i+1] = a7;
         a8 += w1 * z8;
         A8[i+1] = a8;
         a9 += w1 * z9;
         A9[i+1] = a9;
         a10 += w1 * z10;
         A10[i+1] = a10;
         a11 += w1 * z11;
         A11[i+1] = a11;
         a0 = A0[i+2];
         a0 += x2 * y0;
         a1 = A1[i+2];
         a1 += x2 * y1;
         a2 = A2[i+2];
         a2 += x2 * y2;
         a3 = A3[i+2];
         a3 += x2 * y3;
         a4 = A4[i+2];
         a4 += x2 * y4;
         a5 = A5[i+2];
         a5 += x2 * y5;
         a6 = A6[i+2];
         a6 += x2 * y6;
         a7 = A7[i+2];
         a7 += x2 * y7;
         a8 = A8[i+2];
         a8 += x2 * y8;
         a9 = A9[i+2];
         a9 += x2 * y9;
         a10 = A10[i+2];
         a10 += x2 * y10;
         a11 = A11[i+2];
         a11 += x2 * y11;
         a0 += w2 * z0;
         A0[i+2] = a0;
         a1 += w2 * z1;
         A1[i+2] = a1;
         a2 += w2 * z2;
         A2[i+2] = a2;
         a3 += w2 * z3;
         A3[i+2] = a3;
         a4 += w2 * z4;
         A4[i+2] = a4;
         a5 += w2 * z5;
         A5[i+2] = a5;
         a6 += w2 * z6;
         A6[i+2] = a6;
         a7 += w2 * z7;
         A7[i+2] = a7;
         a8 += w2 * z8;
         A8[i+2] = a8;
         a9 += w2 * z9;
         A9[i+2] = a9;
         a10 += w2 * z10;
         A10[i+2] = a10;
         a11 += w2 * z11;
         A11[i+2] = a11;
         a0 = A0[i+3];
         a0 += x3 * y0;
         a1 = A1[i+3];
         a1 += x3 * y1;
         a2 = A2[i+3];
         a2 += x3 * y2;
         a3 = A3[i+3];
         a3 += x3 * y3;
         a4 = A4[i+3];
         a4 += x3 * y4;
         a5 = A5[i+3];
         a5 += x3 * y5;
         a6 = A6[i+3];
         a6 += x3 * y6;
         a7 = A7[i+3];
         a7 += x3 * y7;
         a8 = A8[i+3];
         a8 += x3 * y8;
         a9 = A9[i+3];
         a9 += x3 * y9;
         a10 = A10[i+3];
         a10 += x3 * y10;
         a11 = A11[i+3];
         a11 += x3 * y11;
         a0 += w3 * z0;
         A0[i+3] = a0;
         a1 += w3 * z1;
         A1[i+3] = a1;
         a2 += w3 * z2;
         A2[i+3] = a2;
         a3 += w3 * z3;
         A3[i+3] = a3;
         a4 += w3 * z4;
         A4[i+3] = a4;
         a5 += w3 * z5;
         A5[i+3] = a5;
         a6 += w3 * z6;
         A6[i+3] = a6;
         a7 += w3 * z7;
         A7[i+3] = a7;
         a8 += w3 * z8;
         A8[i+3] = a8;
         a9 += w3 * z9;
         A9[i+3] = a9;
         a10 += w3 * z10;
         A10[i+3] = a10;
         a11 += w3 * z11;
         A11[i+3] = a11;
         a0 = A0[i+4];
         a0 += x4 * y0;
         a1 = A1[i+4];
         a1 += x4 * y1;
         a2 = A2[i+4];
         a2 += x4 * y2;
         a3 = A3[i+4];
         a3 += x4 * y3;
         a4 = A4[i+4];
         a4 += x4 * y4;
         a5 = A5[i+4];
         a5 += x4 * y5;
         a6 = A6[i+4];
         a6 += x4 * y6;
         a7 = A7[i+4];
         a7 += x4 * y7;
         a8 = A8[i+4];
         a8 += x4 * y8;
         a9 = A9[i+4];
         a9 += x4 * y9;
         a10 = A10[i+4];
         a10 += x4 * y10;
         a11 = A11[i+4];
         a11 += x4 * y11;
         a0 += w4 * z0;
         A0[i+4] = a0;
         a1 += w4 * z1;
         A1[i+4] = a1;
         a2 += w4 * z2;
         A2[i+4] = a2;
         a3 += w4 * z3;
         A3[i+4] = a3;
         a4 += w4 * z4;
         A4[i+4] = a4;
         a5 += w4 * z5;
         A5[i+4] = a5;
         a6 += w4 * z6;
         A6[i+4] = a6;
         a7 += w4 * z7;
         A7[i+4] = a7;
         a8 += w4 * z8;
         A8[i+4] = a8;
         a9 += w4 * z9;
         A9[i+4] = a9;
         a10 += w4 * z10;
         A10[i+4] = a10;
         a11 += w4 * z11;
         A11[i+4] = a11;
         a0 = A0[i+5];
         a0 += x5 * y0;
         a1 = A1[i+5];
         a1 += x5 * y1;
         a2 = A2[i+5];
         a2 += x5 * y2;
         a3 = A3[i+5];
         a3 += x5 * y3;
         a4 = A4[i+5];
         a4 += x5 * y4;
         a5 = A5[i+5];
         a5 += x5 * y5;
         a6 = A6[i+5];
         a6 += x5 * y6;
         a7 = A7[i+5];
         a7 += x5 * y7;
         a8 = A8[i+5];
         a8 += x5 * y8;
         a9 = A9[i+5];
         a9 += x5 * y9;
         a10 = A10[i+5];
         a10 += x5 * y10;
         a11 = A11[i+5];
         a11 += x5 * y11;
         a0 += w5 * z0;
         A0[i+5] = a0;
         a1 += w5 * z1;
         A1[i+5] = a1;
         a2 += w5 * z2;
         A2[i+5] = a2;
         a3 += w5 * z3;
         A3[i+5] = a3;
         a4 += w5 * z4;
         A4[i+5] = a4;
         a5 += w5 * z5;
         A5[i+5] = a5;
         a6 += w5 * z6;
         A6[i+5] = a6;
         a7 += w5 * z7;
         A7[i+5] = a7;
         a8 += w5 * z8;
         A8[i+5] = a8;
         a9 += w5 * z9;
         A9[i+5] = a9;
         a10 += w5 * z10;
         A10[i+5] = a10;
         a11 += w5 * z11;
         A11[i+5] = a11;
         a0 = A0[i+6];
         a0 += x6 * y0;
         a1 = A1[i+6];
         a1 += x6 * y1;
         a2 = A2[i+6];
         a2 += x6 * y2;
         a3 = A3[i+6];
         a3 += x6 * y3;
         a4 = A4[i+6];
         a4 += x6 * y4;
         a5 = A5[i+6];
         a5 += x6 * y5;
         a6 = A6[i+6];
         a6 += x6 * y6;
         a7 = A7[i+6];
         a7 += x6 * y7;
         a8 = A8[i+6];
         a8 += x6 * y8;
         a9 = A9[i+6];
         a9 += x6 * y9;
         a10 = A10[i+6];
         a10 += x6 * y10;
         a11 = A11[i+6];
         a11 += x6 * y11;
         a0 += w6 * z0;
         A0[i+6] = a0;
         a1 += w6 * z1;
         A1[i+6] = a1;
         a2 += w6 * z2;
         A2[i+6] = a2;
         a3 += w6 * z3;
         A3[i+6] = a3;
         a4 += w6 * z4;
         A4[i+6] = a4;
         a5 += w6 * z5;
         A5[i+6] = a5;
         a6 += w6 * z6;
         A6[i+6] = a6;
         a7 += w6 * z7;
         A7[i+6] = a7;
         a8 += w6 * z8;
         A8[i+6] = a8;
         a9 += w6 * z9;
         A9[i+6] = a9;
         a10 += w6 * z10;
         A10[i+6] = a10;
         a11 += w6 * z11;
         A11[i+6] = a11;
         a0 = A0[i+7];
         a0 += x7 * y0;
         a1 = A1[i+7];
         a1 += x7 * y1;
         a2 = A2[i+7];
         a2 += x7 * y2;
         a3 = A3[i+7];
         a3 += x7 * y3;
         a4 = A4[i+7];
         a4 += x7 * y4;
         a5 = A5[i+7];
         a5 += x7 * y5;
         a6 = A6[i+7];
         a6 += x7 * y6;
         a7 = A7[i+7];
         a7 += x7 * y7;
         a8 = A8[i+7];
         a8 += x7 * y8;
         a9 = A9[i+7];
         a9 += x7 * y9;
         a10 = A10[i+7];
         a10 += x7 * y10;
         a11 = A11[i+7];
         a11 += x7 * y11;
         a0 += w7 * z0;
         A0[i+7] = a0;
         a1 += w7 * z1;
         A1[i+7] = a1;
         a2 += w7 * z2;
         A2[i+7] = a2;
         a3 += w7 * z3;
         A3[i+7] = a3;
         a4 += w7 * z4;
         A4[i+7] = a4;
         a5 += w7 * z5;
         A5[i+7] = a5;
         a6 += w7 * z6;
         A6[i+7] = a6;
         a7 += w7 * z7;
         A7[i+7] = a7;
         a8 += w7 * z8;
         A8[i+7] = a8;
         a9 += w7 * z9;
         A9[i+7] = a9;
         a10 += w7 * z10;
         A10[i+7] = a10;
         a11 += w7 * z11;
         A11[i+7] = a11;
         a0 = A0[i+8];
         a0 += x8 * y0;
         a1 = A1[i+8];
         a1 += x8 * y1;
         a2 = A2[i+8];
         a2 += x8 * y2;
         a3 = A3[i+8];
         a3 += x8 * y3;
         a4 = A4[i+8];
         a4 += x8 * y4;
         a5 = A5[i+8];
         a5 += x8 * y5;
         a6 = A6[i+8];
         a6 += x8 * y6;
         a7 = A7[i+8];
         a7 += x8 * y7;
         a8 = A8[i+8];
         a8 += x8 * y8;
         a9 = A9[i+8];
         a9 += x8 * y9;
         a10 = A10[i+8];
         a10 += x8 * y10;
         a11 = A11[i+8];
         a11 += x8 * y11;
         a0 += w8 * z0;
         prefA(A0);
         A0[i+8] = a0;
         a1 += w8 * z1;
         prefA(A1);
         A1[i+8] = a1;
         a2 += w8 * z2;
         prefA(A2);
         A2[i+8] = a2;
         a3 += w8 * z3;
         prefA(A3);
         A3[i+8] = a3;
         a4 += w8 * z4;
         prefA(A4);
         A4[i+8] = a4;
         a5 += w8 * z5;
         prefA(A5);
         A5[i+8] = a5;
         a6 += w8 * z6;
         prefA(A6);
         A6[i+8] = a6;
         a7 += w8 * z7;
         prefA(A7);
         A7[i+8] = a7;
         a8 += w8 * z8;
         prefA(A8);
         A8[i+8] = a8;
         a9 += w8 * z9;
         prefA(A9);
         A9[i+8] = a9;
         a10 += w8 * z10;
         prefA(A10);
         A10[i+8] = a10;
         a11 += w8 * z11;
         prefA(A11);
         A11[i+8] = a11;
         a0 = A0[i+9];
         a0 += x9 * y0;
         a1 = A1[i+9];
         a1 += x9 * y1;
         a2 = A2[i+9];
         a2 += x9 * y2;
         a3 = A3[i+9];
         a3 += x9 * y3;
         a4 = A4[i+9];
         a4 += x9 * y4;
         a5 = A5[i+9];
         a5 += x9 * y5;
         a6 = A6[i+9];
         a6 += x9 * y6;
         a7 = A7[i+9];
         a7 += x9 * y7;
         a8 = A8[i+9];
         a8 += x9 * y8;
         a9 = A9[i+9];
         a9 += x9 * y9;
         a10 = A10[i+9];
         a10 += x9 * y10;
         a11 = A11[i+9];
         a11 += x9 * y11;
         a0 += w9 * z0;
         A0[i+9] = a0;
         a1 += w9 * z1;
         A1[i+9] = a1;
         a2 += w9 * z2;
         A2[i+9] = a2;
         a3 += w9 * z3;
         A3[i+9] = a3;
         a4 += w9 * z4;
         A4[i+9] = a4;
         a5 += w9 * z5;
         A5[i+9] = a5;
         a6 += w9 * z6;
         A6[i+9] = a6;
         a7 += w9 * z7;
         A7[i+9] = a7;
         a8 += w9 * z8;
         A8[i+9] = a8;
         a9 += w9 * z9;
         A9[i+9] = a9;
         a10 += w9 * z10;
         A10[i+9] = a10;
         a11 += w9 * z11;
         A11[i+9] = a11;
         a0 = A0[i+10];
         a0 += x10 * y0;
         a1 = A1[i+10];
         a1 += x10 * y1;
         a2 = A2[i+10];
         a2 += x10 * y2;
         a3 = A3[i+10];
         a3 += x10 * y3;
         a4 = A4[i+10];
         a4 += x10 * y4;
         a5 = A5[i+10];
         a5 += x10 * y5;
         a6 = A6[i+10];
         a6 += x10 * y6;
         a7 = A7[i+10];
         a7 += x10 * y7;
         a8 = A8[i+10];
         a8 += x10 * y8;
         a9 = A9[i+10];
         a9 += x10 * y9;
         a10 = A10[i+10];
         a10 += x10 * y10;
         a11 = A11[i+10];
         a11 += x10 * y11;
         a0 += w10 * z0;
         A0[i+10] = a0;
         a1 += w10 * z1;
         A1[i+10] = a1;
         a2 += w10 * z2;
         A2[i+10] = a2;
         a3 += w10 * z3;
         A3[i+10] = a3;
         a4 += w10 * z4;
         A4[i+10] = a4;
         a5 += w10 * z5;
         A5[i+10] = a5;
         a6 += w10 * z6;
         A6[i+10] = a6;
         a7 += w10 * z7;
         A7[i+10] = a7;
         a8 += w10 * z8;
         A8[i+10] = a8;
         a9 += w10 * z9;
         A9[i+10] = a9;
         a10 += w10 * z10;
         A10[i+10] = a10;
         a11 += w10 * z11;
         A11[i+10] = a11;
         a0 = A0[i+11];
         a0 += x11 * y0;
         a1 = A1[i+11];
         a1 += x11 * y1;
         a2 = A2[i+11];
         a2 += x11 * y2;
         a3 = A3[i+11];
         a3 += x11 * y3;
         a4 = A4[i+11];
         a4 += x11 * y4;
         a5 = A5[i+11];
         a5 += x11 * y5;
         a6 = A6[i+11];
         a6 += x11 * y6;
         a7 = A7[i+11];
         a7 += x11 * y7;
         a8 = A8[i+11];
         a8 += x11 * y8;
         a9 = A9[i+11];
         a9 += x11 * y9;
         a10 = A10[i+11];
         a10 += x11 * y10;
         a11 = A11[i+11];
         a11 += x11 * y11;
         a0 += w11 * z0;
         A0[i+11] = a0;
         a1 += w11 * z1;
         A1[i+11] = a1;
         a2 += w11 * z2;
         A2[i+11] = a2;
         a3 += w11 * z3;
         A3[i+11] = a3;
         a4 += w11 * z4;
         A4[i+11] = a4;
         a5 += w11 * z5;
         A5[i+11] = a5;
         a6 += w11 * z6;
         A6[i+11] = a6;
         a7 += w11 * z7;
         A7[i+11] = a7;
         a8 += w11 * z8;
         A8[i+11] = a8;
         a9 += w11 * z9;
         A9[i+11] = a9;
         a10 += w11 * z10;
         A10[i+11] = a10;
         a11 += w11 * z11;
         A11[i+11] = a11;
         a0 = A0[i+12];
         a0 += x12 * y0;
         a1 = A1[i+12];
         a1 += x12 * y1;
         a2 = A2[i+12];
         a2 += x12 * y2;
         a3 = A3[i+12];
         a3 += x12 * y3;
         a4 = A4[i+12];
         a4 += x12 * y4;
         a5 = A5[i+12];
         a5 += x12 * y5;
         a6 = A6[i+12];
         a6 += x12 * y6;
         a7 = A7[i+12];
         a7 += x12 * y7;
         a8 = A8[i+12];
         a8 += x12 * y8;
         a9 = A9[i+12];
         a9 += x12 * y9;
         a10 = A10[i+12];
         a10 += x12 * y10;
         a11 = A11[i+12];
         a11 += x12 * y11;
         a0 += w12 * z0;
         A0[i+12] = a0;
         a1 += w12 * z1;
         A1[i+12] = a1;
         a2 += w12 * z2;
         A2[i+12] = a2;
         a3 += w12 * z3;
         A3[i+12] = a3;
         a4 += w12 * z4;
         A4[i+12] = a4;
         a5 += w12 * z5;
         A5[i+12] = a5;
         a6 += w12 * z6;
         A6[i+12] = a6;
         a7 += w12 * z7;
         A7[i+12] = a7;
         a8 += w12 * z8;
         A8[i+12] = a8;
         a9 += w12 * z9;
         A9[i+12] = a9;
         a10 += w12 * z10;
         A10[i+12] = a10;
         a11 += w12 * z11;
         A11[i+12] = a11;
         a0 = A0[i+13];
         a0 += x13 * y0;
         a1 = A1[i+13];
         a1 += x13 * y1;
         a2 = A2[i+13];
         a2 += x13 * y2;
         a3 = A3[i+13];
         a3 += x13 * y3;
         a4 = A4[i+13];
         a4 += x13 * y4;
         a5 = A5[i+13];
         a5 += x13 * y5;
         a6 = A6[i+13];
         a6 += x13 * y6;
         a7 = A7[i+13];
         a7 += x13 * y7;
         a8 = A8[i+13];
         a8 += x13 * y8;
         a9 = A9[i+13];
         a9 += x13 * y9;
         a10 = A10[i+13];
         a10 += x13 * y10;
         a11 = A11[i+13];
         a11 += x13 * y11;
         a0 += w13 * z0;
         A0[i+13] = a0;
         a1 += w13 * z1;
         A1[i+13] = a1;
         a2 += w13 * z2;
         A2[i+13] = a2;
         a3 += w13 * z3;
         A3[i+13] = a3;
         a4 += w13 * z4;
         A4[i+13] = a4;
         a5 += w13 * z5;
         A5[i+13] = a5;
         a6 += w13 * z6;
         A6[i+13] = a6;
         a7 += w13 * z7;
         A7[i+13] = a7;
         a8 += w13 * z8;
         A8[i+13] = a8;
         a9 += w13 * z9;
         A9[i+13] = a9;
         a10 += w13 * z10;
         A10[i+13] = a10;
         a11 += w13 * z11;
         A11[i+13] = a11;
         a0 = A0[i+14];
         a0 += x14 * y0;
         a1 = A1[i+14];
         a1 += x14 * y1;
         a2 = A2[i+14];
         a2 += x14 * y2;
         a3 = A3[i+14];
         a3 += x14 * y3;
         a4 = A4[i+14];
         a4 += x14 * y4;
         a5 = A5[i+14];
         a5 += x14 * y5;
         a6 = A6[i+14];
         a6 += x14 * y6;
         a7 = A7[i+14];
         a7 += x14 * y7;
         a8 = A8[i+14];
         a8 += x14 * y8;
         a9 = A9[i+14];
         a9 += x14 * y9;
         a10 = A10[i+14];
         a10 += x14 * y10;
         a11 = A11[i+14];
         a11 += x14 * y11;
         a0 += w14 * z0;
         A0[i+14] = a0;
         a1 += w14 * z1;
         A1[i+14] = a1;
         a2 += w14 * z2;
         A2[i+14] = a2;
         a3 += w14 * z3;
         A3[i+14] = a3;
         a4 += w14 * z4;
         A4[i+14] = a4;
         a5 += w14 * z5;
         A5[i+14] = a5;
         a6 += w14 * z6;
         A6[i+14] = a6;
         a7 += w14 * z7;
         A7[i+14] = a7;
         a8 += w14 * z8;
         A8[i+14] = a8;
         a9 += w14 * z9;
         A9[i+14] = a9;
         a10 += w14 * z10;
         A10[i+14] = a10;
         a11 += w14 * z11;
         A11[i+14] = a11;
         a0 = A0[i+15];
         a0 += x15 * y0;
         a1 = A1[i+15];
         a1 += x15 * y1;
         a2 = A2[i+15];
         a2 += x15 * y2;
         a3 = A3[i+15];
         a3 += x15 * y3;
         a4 = A4[i+15];
         a4 += x15 * y4;
         a5 = A5[i+15];
         a5 += x15 * y5;
         a6 = A6[i+15];
         a6 += x15 * y6;
         a7 = A7[i+15];
         a7 += x15 * y7;
         a8 = A8[i+15];
         a8 += x15 * y8;
         a9 = A9[i+15];
         a9 += x15 * y9;
         a10 = A10[i+15];
         a10 += x15 * y10;
         a11 = A11[i+15];
         a11 += x15 * y11;
         a0 += w15 * z0;
         A0[i+15] = a0;
         a1 += w15 * z1;
         A1[i+15] = a1;
         a2 += w15 * z2;
         A2[i+15] = a2;
         a3 += w15 * z3;
         A3[i+15] = a3;
         a4 += w15 * z4;
         A4[i+15] = a4;
         a5 += w15 * z5;
         A5[i+15] = a5;
         a6 += w15 * z6;
         A6[i+15] = a6;
         a7 += w15 * z7;
         A7[i+15] = a7;
         a8 += w15 * z8;
         A8[i+15] = a8;
         a9 += w15 * z9;
         A9[i+15] = a9;
         a10 += w15 * z10;
         A10[i+15] = a10;
         a11 += w15 * z11;
         A11[i+15] = a11;
         a0 = A0[i+16];
         a0 += x16 * y0;
         a1 = A1[i+16];
         a1 += x16 * y1;
         a2 = A2[i+16];
         a2 += x16 * y2;
         a3 = A3[i+16];
         a3 += x16 * y3;
         a4 = A4[i+16];
         a4 += x16 * y4;
         a5 = A5[i+16];
         a5 += x16 * y5;
         a6 = A6[i+16];
         a6 += x16 * y6;
         a7 = A7[i+16];
         a7 += x16 * y7;
         a8 = A8[i+16];
         a8 += x16 * y8;
         a9 = A9[i+16];
         a9 += x16 * y9;
         a10 = A10[i+16];
         a10 += x16 * y10;
         a11 = A11[i+16];
         a11 += x16 * y11;
         a0 += w16 * z0;
         prefA(A0);
         A0[i+16] = a0;
         a1 += w16 * z1;
         prefA(A1);
         A1[i+16] = a1;
         a2 += w16 * z2;
         prefA(A2);
         A2[i+16] = a2;
         a3 += w16 * z3;
         prefA(A3);
         A3[i+16] = a3;
         a4 += w16 * z4;
         prefA(A4);
         A4[i+16] = a4;
         a5 += w16 * z5;
         prefA(A5);
         A5[i+16] = a5;
         a6 += w16 * z6;
         prefA(A6);
         A6[i+16] = a6;
         a7 += w16 * z7;
         prefA(A7);
         A7[i+16] = a7;
         a8 += w16 * z8;
         prefA(A8);
         A8[i+16] = a8;
         a9 += w16 * z9;
         prefA(A9);
         A9[i+16] = a9;
         a10 += w16 * z10;
         prefA(A10);
         A10[i+16] = a10;
         a11 += w16 * z11;
         prefA(A11);
         A11[i+16] = a11;
         a0 = A0[i+17];
         a0 += x17 * y0;
         a1 = A1[i+17];
         a1 += x17 * y1;
         a2 = A2[i+17];
         a2 += x17 * y2;
         a3 = A3[i+17];
         a3 += x17 * y3;
         a4 = A4[i+17];
         a4 += x17 * y4;
         a5 = A5[i+17];
         a5 += x17 * y5;
         a6 = A6[i+17];
         a6 += x17 * y6;
         a7 = A7[i+17];
         a7 += x17 * y7;
         a8 = A8[i+17];
         a8 += x17 * y8;
         a9 = A9[i+17];
         a9 += x17 * y9;
         a10 = A10[i+17];
         a10 += x17 * y10;
         a11 = A11[i+17];
         a11 += x17 * y11;
         a0 += w17 * z0;
         A0[i+17] = a0;
         a1 += w17 * z1;
         A1[i+17] = a1;
         a2 += w17 * z2;
         A2[i+17] = a2;
         a3 += w17 * z3;
         A3[i+17] = a3;
         a4 += w17 * z4;
         A4[i+17] = a4;
         a5 += w17 * z5;
         A5[i+17] = a5;
         a6 += w17 * z6;
         A6[i+17] = a6;
         a7 += w17 * z7;
         A7[i+17] = a7;
         a8 += w17 * z8;
         A8[i+17] = a8;
         a9 += w17 * z9;
         A9[i+17] = a9;
         a10 += w17 * z10;
         A10[i+17] = a10;
         a11 += w17 * z11;
         A11[i+17] = a11;
         a0 = A0[i+18];
         a0 += x18 * y0;
         a1 = A1[i+18];
         a1 += x18 * y1;
         a2 = A2[i+18];
         a2 += x18 * y2;
         a3 = A3[i+18];
         a3 += x18 * y3;
         a4 = A4[i+18];
         a4 += x18 * y4;
         a5 = A5[i+18];
         a5 += x18 * y5;
         a6 = A6[i+18];
         a6 += x18 * y6;
         a7 = A7[i+18];
         a7 += x18 * y7;
         a8 = A8[i+18];
         a8 += x18 * y8;
         a9 = A9[i+18];
         a9 += x18 * y9;
         a10 = A10[i+18];
         a10 += x18 * y10;
         a11 = A11[i+18];
         a11 += x18 * y11;
         a0 += w18 * z0;
         A0[i+18] = a0;
         a1 += w18 * z1;
         A1[i+18] = a1;
         a2 += w18 * z2;
         A2[i+18] = a2;
         a3 += w18 * z3;
         A3[i+18] = a3;
         a4 += w18 * z4;
         A4[i+18] = a4;
         a5 += w18 * z5;
         A5[i+18] = a5;
         a6 += w18 * z6;
         A6[i+18] = a6;
         a7 += w18 * z7;
         A7[i+18] = a7;
         a8 += w18 * z8;
         A8[i+18] = a8;
         a9 += w18 * z9;
         A9[i+18] = a9;
         a10 += w18 * z10;
         A10[i+18] = a10;
         a11 += w18 * z11;
         A11[i+18] = a11;
         a0 = A0[i+19];
         a0 += x19 * y0;
         a1 = A1[i+19];
         a1 += x19 * y1;
         a2 = A2[i+19];
         a2 += x19 * y2;
         a3 = A3[i+19];
         a3 += x19 * y3;
         a4 = A4[i+19];
         a4 += x19 * y4;
         a5 = A5[i+19];
         a5 += x19 * y5;
         a6 = A6[i+19];
         a6 += x19 * y6;
         a7 = A7[i+19];
         a7 += x19 * y7;
         a8 = A8[i+19];
         a8 += x19 * y8;
         a9 = A9[i+19];
         a9 += x19 * y9;
         a10 = A10[i+19];
         a10 += x19 * y10;
         a11 = A11[i+19];
         a11 += x19 * y11;
         a0 += w19 * z0;
         A0[i+19] = a0;
         a1 += w19 * z1;
         A1[i+19] = a1;
         a2 += w19 * z2;
         A2[i+19] = a2;
         a3 += w19 * z3;
         A3[i+19] = a3;
         a4 += w19 * z4;
         A4[i+19] = a4;
         a5 += w19 * z5;
         A5[i+19] = a5;
         a6 += w19 * z6;
         A6[i+19] = a6;
         a7 += w19 * z7;
         A7[i+19] = a7;
         a8 += w19 * z8;
         A8[i+19] = a8;
         a9 += w19 * z9;
         A9[i+19] = a9;
         a10 += w19 * z10;
         A10[i+19] = a10;
         a11 += w19 * z11;
         A11[i+19] = a11;
         a0 = A0[i+20];
         a0 += x20 * y0;
         a1 = A1[i+20];
         a1 += x20 * y1;
         a2 = A2[i+20];
         a2 += x20 * y2;
         a3 = A3[i+20];
         a3 += x20 * y3;
         a4 = A4[i+20];
         a4 += x20 * y4;
         a5 = A5[i+20];
         a5 += x20 * y5;
         a6 = A6[i+20];
         a6 += x20 * y6;
         a7 = A7[i+20];
         a7 += x20 * y7;
         a8 = A8[i+20];
         a8 += x20 * y8;
         a9 = A9[i+20];
         a9 += x20 * y9;
         a10 = A10[i+20];
         a10 += x20 * y10;
         a11 = A11[i+20];
         a11 += x20 * y11;
         a0 += w20 * z0;
         A0[i+20] = a0;
         a1 += w20 * z1;
         A1[i+20] = a1;
         a2 += w20 * z2;
         A2[i+20] = a2;
         a3 += w20 * z3;
         A3[i+20] = a3;
         a4 += w20 * z4;
         A4[i+20] = a4;
         a5 += w20 * z5;
         A5[i+20] = a5;
         a6 += w20 * z6;
         A6[i+20] = a6;
         a7 += w20 * z7;
         A7[i+20] = a7;
         a8 += w20 * z8;
         A8[i+20] = a8;
         a9 += w20 * z9;
         A9[i+20] = a9;
         a10 += w20 * z10;
         A10[i+20] = a10;
         a11 += w20 * z11;
         A11[i+20] = a11;
         a0 = A0[i+21];
         a0 += x21 * y0;
         a1 = A1[i+21];
         a1 += x21 * y1;
         a2 = A2[i+21];
         a2 += x21 * y2;
         a3 = A3[i+21];
         a3 += x21 * y3;
         a4 = A4[i+21];
         a4 += x21 * y4;
         a5 = A5[i+21];
         a5 += x21 * y5;
         a6 = A6[i+21];
         a6 += x21 * y6;
         a7 = A7[i+21];
         a7 += x21 * y7;
         a8 = A8[i+21];
         a8 += x21 * y8;
         a9 = A9[i+21];
         a9 += x21 * y9;
         a10 = A10[i+21];
         a10 += x21 * y10;
         a11 = A11[i+21];
         a11 += x21 * y11;
         a0 += w21 * z0;
         A0[i+21] = a0;
         a1 += w21 * z1;
         A1[i+21] = a1;
         a2 += w21 * z2;
         A2[i+21] = a2;
         a3 += w21 * z3;
         A3[i+21] = a3;
         a4 += w21 * z4;
         A4[i+21] = a4;
         a5 += w21 * z5;
         A5[i+21] = a5;
         a6 += w21 * z6;
         A6[i+21] = a6;
         a7 += w21 * z7;
         A7[i+21] = a7;
         a8 += w21 * z8;
         A8[i+21] = a8;
         a9 += w21 * z9;
         A9[i+21] = a9;
         a10 += w21 * z10;
         A10[i+21] = a10;
         a11 += w21 * z11;
         A11[i+21] = a11;
         a0 = A0[i+22];
         a0 += x22 * y0;
         a1 = A1[i+22];
         a1 += x22 * y1;
         a2 = A2[i+22];
         a2 += x22 * y2;
         a3 = A3[i+22];
         a3 += x22 * y3;
         a4 = A4[i+22];
         a4 += x22 * y4;
         a5 = A5[i+22];
         a5 += x22 * y5;
         a6 = A6[i+22];
         a6 += x22 * y6;
         a7 = A7[i+22];
         a7 += x22 * y7;
         a8 = A8[i+22];
         a8 += x22 * y8;
         a9 = A9[i+22];
         a9 += x22 * y9;
         a10 = A10[i+22];
         a10 += x22 * y10;
         a11 = A11[i+22];
         a11 += x22 * y11;
         a0 += w22 * z0;
         A0[i+22] = a0;
         a1 += w22 * z1;
         A1[i+22] = a1;
         a2 += w22 * z2;
         A2[i+22] = a2;
         a3 += w22 * z3;
         A3[i+22] = a3;
         a4 += w22 * z4;
         A4[i+22] = a4;
         a5 += w22 * z5;
         A5[i+22] = a5;
         a6 += w22 * z6;
         A6[i+22] = a6;
         a7 += w22 * z7;
         A7[i+22] = a7;
         a8 += w22 * z8;
         A8[i+22] = a8;
         a9 += w22 * z9;
         A9[i+22] = a9;
         a10 += w22 * z10;
         A10[i+22] = a10;
         a11 += w22 * z11;
         A11[i+22] = a11;
         a0 = A0[i+23];
         a0 += x23 * y0;
         a1 = A1[i+23];
         a1 += x23 * y1;
         a2 = A2[i+23];
         a2 += x23 * y2;
         a3 = A3[i+23];
         a3 += x23 * y3;
         a4 = A4[i+23];
         a4 += x23 * y4;
         a5 = A5[i+23];
         a5 += x23 * y5;
         a6 = A6[i+23];
         a6 += x23 * y6;
         a7 = A7[i+23];
         a7 += x23 * y7;
         a8 = A8[i+23];
         a8 += x23 * y8;
         a9 = A9[i+23];
         a9 += x23 * y9;
         a10 = A10[i+23];
         a10 += x23 * y10;
         a11 = A11[i+23];
         a11 += x23 * y11;
         a0 += w23 * z0;
         A0[i+23] = a0;
         a1 += w23 * z1;
         A1[i+23] = a1;
         a2 += w23 * z2;
         A2[i+23] = a2;
         a3 += w23 * z3;
         A3[i+23] = a3;
         a4 += w23 * z4;
         A4[i+23] = a4;
         a5 += w23 * z5;
         A5[i+23] = a5;
         a6 += w23 * z6;
         A6[i+23] = a6;
         a7 += w23 * z7;
         A7[i+23] = a7;
         a8 += w23 * z8;
         A8[i+23] = a8;
         a9 += w23 * z9;
         A9[i+23] = a9;
         a10 += w23 * z10;
         A10[i+23] = a10;
         a11 += w23 * z11;
         A11[i+23] = a11;

      }
      for (i=M24; i < M; i++)
      {
         const register double x0 = X[i], w0 = W[i];
         register double a0=A0[i], a1=A1[i], a2=A2[i], a3=A3[i], a4=A4[i],
                         a5=A5[i], a6=A6[i], a7=A7[i], a8=A8[i], a9=A9[i],
                         a10=A10[i], a11=A11[i];

         a0 += x0 * y0;
         a1 += x0 * y1;
         a2 += x0 * y2;
         a3 += x0 * y3;
         a4 += x0 * y4;
         a5 += x0 * y5;
         a6 += x0 * y6;
         a7 += x0 * y7;
         a8 += x0 * y8;
         a9 += x0 * y9;
         a10 += x0 * y10;
         a11 += x0 * y11;
         a0 += w0 * z0;
         A0[i] = a0;
         a1 += w0 * z1;
         A1[i] = a1;
         a2 += w0 * z2;
         A2[i] = a2;
         a3 += w0 * z3;
         A3[i] = a3;
         a4 += w0 * z4;
         A4[i] = a4;
         a5 += w0 * z5;
         A5[i] = a5;
         a6 += w0 * z6;
         A6[i] = a6;
         a7 += w0 * z7;
         A7[i] = a7;
         a8 += w0 * z8;
         A8[i] = a8;
         a9 += w0 * z9;
         A9[i] = a9;
         a10 += w0 * z10;
         A10[i] = a10;
         a11 += w0 * z11;
         A11[i] = a11;
      }
      Z += 12;
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N12; j; j--, A += lda, Y++)
   {
      double *A0=A;
      const register double y0=Y[0], z0=Z[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      prefY(Z+1+1-1);
      for (i=0; i < M24; i += 24)
      {
         register double a0;
         const double x0=X[i+0], w0=W[i+0], x1=X[i+1], w1=W[i+1], x2=X[i+2],
                      w2=W[i+2], x3=X[i+3], w3=W[i+3], x4=X[i+4], w4=W[i+4],
                      x5=X[i+5], w5=W[i+5], x6=X[i+6], w6=W[i+6], x7=X[i+7],
                      w7=W[i+7], x8=X[i+8], w8=W[i+8], x9=X[i+9], w9=W[i+9],
                      x10=X[i+10], w10=W[i+10], x11=X[i+11], w11=W[i+11],
                      x12=X[i+12], w12=W[i+12], x13=X[i+13], w13=W[i+13],
                      x14=X[i+14], w14=W[i+14], x15=X[i+15], w15=W[i+15],
                      x16=X[i+16], w16=W[i+16], x17=X[i+17], w17=W[i+17],
                      x18=X[i+18], w18=W[i+18], x19=X[i+19], w19=W[i+19],
                      x20=X[i+20], w20=W[i+20], x21=X[i+21], w21=W[i+21],
                      x22=X[i+22], w22=W[i+22], x23=X[i+23], w23=W[i+23];

         prefX(X);
         a0 = A0[i+0];
         a0 += x0 * y0;
         prefX(W);
         a0 += w0 * z0;
         prefA(A0);
         A0[i+0] = a0;
         a0 = A0[i+1];
         a0 += x1 * y0;
         a0 += w1 * z0;
         A0[i+1] = a0;
         a0 = A0[i+2];
         a0 += x2 * y0;
         a0 += w2 * z0;
         A0[i+2] = a0;
         a0 = A0[i+3];
         a0 += x3 * y0;
         a0 += w3 * z0;
         A0[i+3] = a0;
         a0 = A0[i+4];
         a0 += x4 * y0;
         a0 += w4 * z0;
         A0[i+4] = a0;
         a0 = A0[i+5];
         a0 += x5 * y0;
         a0 += w5 * z0;
         A0[i+5] = a0;
         a0 = A0[i+6];
         a0 += x6 * y0;
         a0 += w6 * z0;
         A0[i+6] = a0;
         a0 = A0[i+7];
         a0 += x7 * y0;
         a0 += w7 * z0;
         A0[i+7] = a0;
         a0 = A0[i+8];
         a0 += x8 * y0;
         a0 += w8 * z0;
         prefA(A0);
         A0[i+8] = a0;
         a0 = A0[i+9];
         a0 += x9 * y0;
         a0 += w9 * z0;
         A0[i+9] = a0;
         a0 = A0[i+10];
         a0 += x10 * y0;
         a0 += w10 * z0;
         A0[i+10] = a0;
         a0 = A0[i+11];
         a0 += x11 * y0;
         a0 += w11 * z0;
         A0[i+11] = a0;
         a0 = A0[i+12];
         a0 += x12 * y0;
         a0 += w12 * z0;
         A0[i+12] = a0;
         a0 = A0[i+13];
         a0 += x13 * y0;
         a0 += w13 * z0;
         A0[i+13] = a0;
         a0 = A0[i+14];
         a0 += x14 * y0;
         a0 += w14 * z0;
         A0[i+14] = a0;
         a0 = A0[i+15];
         a0 += x15 * y0;
         a0 += w15 * z0;
         A0[i+15] = a0;
         a0 = A0[i+16];
         a0 += x16 * y0;
         a0 += w16 * z0;
         prefA(A0);
         A0[i+16] = a0;
         a0 = A0[i+17];
         a0 += x17 * y0;
         a0 += w17 * z0;
         A0[i+17] = a0;
         a0 = A0[i+18];
         a0 += x18 * y0;
         a0 += w18 * z0;
         A0[i+18] = a0;
         a0 = A0[i+19];
         a0 += x19 * y0;
         a0 += w19 * z0;
         A0[i+19] = a0;
         a0 = A0[i+20];
         a0 += x20 * y0;
         a0 += w20 * z0;
         A0[i+20] = a0;
         a0 = A0[i+21];
         a0 += x21 * y0;
         a0 += w21 * z0;
         A0[i+21] = a0;
         a0 = A0[i+22];
         a0 += x22 * y0;
         a0 += w22 * z0;
         A0[i+22] = a0;
         a0 = A0[i+23];
         a0 += x23 * y0;
         a0 += w23 * z0;
         A0[i+23] = a0;

      }
      for (i=M24; i < M; i++)
      {
         const register double x0 = X[i], w0 = W[i];
         register double a0=A0[i];

         a0 += x0 * y0;
         a0 += w0 * z0;
         A0[i] = a0;
      }
      Z++;
   }
}
