#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 4
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M0, ATL_CINT N, const double *X, const double *Y, 
               double *A, ATL_CINT lda0)
/*
 * 32x4 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N4=(N/4)*4, M=M0+M0, M32=((M0/32)*32)<<1,
            lda=lda0+lda0, lda4=lda*4;
   ATL_INT j;

   for (j=N4; j; j -= 4, A += lda4, Y += 8)
   {
      double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda;
      const register double ry0=Y[0], iy0=Y[1], ry1=Y[2], iy1=Y[3], ry2=Y[4],
                            iy2=Y[5], ry3=Y[6], iy3=Y[7];
      ATL_INT i;
      prefY(Y+8+8-1);
      for (i=0; i < M32; i += 64)
      {
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31], rx16=X[i+32],
                               ix16=X[i+33], rx17=X[i+34], ix17=X[i+35],
                               rx18=X[i+36], ix18=X[i+37], rx19=X[i+38],
                               ix19=X[i+39], rx20=X[i+40], ix20=X[i+41],
                               rx21=X[i+42], ix21=X[i+43], rx22=X[i+44],
                               ix22=X[i+45], rx23=X[i+46], ix23=X[i+47],
                               rx24=X[i+48], ix24=X[i+49], rx25=X[i+50],
                               ix25=X[i+51], rx26=X[i+52], ix26=X[i+53],
                               rx27=X[i+54], ix27=X[i+55], rx28=X[i+56],
                               ix28=X[i+57], rx29=X[i+58], ix29=X[i+59],
                               rx30=X[i+60], ix30=X[i+61], rx31=X[i+62],
                               ix31=X[i+63];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A1[i+0] += ry1*rx0 - iy1*ix0;
         prefA(A1);
         A1[i+0+1] += ry1*ix0 + iy1*rx0;
         A2[i+0] += ry2*rx0 - iy2*ix0;
         prefA(A2);
         A2[i+0+1] += ry2*ix0 + iy2*rx0;
         A3[i+0] += ry3*rx0 - iy3*ix0;
         prefA(A3);
         A3[i+0+1] += ry3*ix0 + iy3*rx0;
         A0[i+2] += ry0*rx1 - iy0*ix1;
         A0[i+2+1] += ry0*ix1 + iy0*rx1;
         A1[i+2] += ry1*rx1 - iy1*ix1;
         A1[i+2+1] += ry1*ix1 + iy1*rx1;
         A2[i+2] += ry2*rx1 - iy2*ix1;
         A2[i+2+1] += ry2*ix1 + iy2*rx1;
         A3[i+2] += ry3*rx1 - iy3*ix1;
         A3[i+2+1] += ry3*ix1 + iy3*rx1;
         A0[i+4] += ry0*rx2 - iy0*ix2;
         A0[i+4+1] += ry0*ix2 + iy0*rx2;
         A1[i+4] += ry1*rx2 - iy1*ix2;
         A1[i+4+1] += ry1*ix2 + iy1*rx2;
         A2[i+4] += ry2*rx2 - iy2*ix2;
         A2[i+4+1] += ry2*ix2 + iy2*rx2;
         A3[i+4] += ry3*rx2 - iy3*ix2;
         A3[i+4+1] += ry3*ix2 + iy3*rx2;
         A0[i+6] += ry0*rx3 - iy0*ix3;
         A0[i+6+1] += ry0*ix3 + iy0*rx3;
         A1[i+6] += ry1*rx3 - iy1*ix3;
         A1[i+6+1] += ry1*ix3 + iy1*rx3;
         A2[i+6] += ry2*rx3 - iy2*ix3;
         A2[i+6+1] += ry2*ix3 + iy2*rx3;
         A3[i+6] += ry3*rx3 - iy3*ix3;
         A3[i+6+1] += ry3*ix3 + iy3*rx3;
         A0[i+8] += ry0*rx4 - iy0*ix4;
         prefA(A0);
         A0[i+8+1] += ry0*ix4 + iy0*rx4;
         A1[i+8] += ry1*rx4 - iy1*ix4;
         prefA(A1);
         A1[i+8+1] += ry1*ix4 + iy1*rx4;
         A2[i+8] += ry2*rx4 - iy2*ix4;
         prefA(A2);
         A2[i+8+1] += ry2*ix4 + iy2*rx4;
         A3[i+8] += ry3*rx4 - iy3*ix4;
         prefA(A3);
         A3[i+8+1] += ry3*ix4 + iy3*rx4;
         A0[i+10] += ry0*rx5 - iy0*ix5;
         A0[i+10+1] += ry0*ix5 + iy0*rx5;
         A1[i+10] += ry1*rx5 - iy1*ix5;
         A1[i+10+1] += ry1*ix5 + iy1*rx5;
         A2[i+10] += ry2*rx5 - iy2*ix5;
         A2[i+10+1] += ry2*ix5 + iy2*rx5;
         A3[i+10] += ry3*rx5 - iy3*ix5;
         A3[i+10+1] += ry3*ix5 + iy3*rx5;
         A0[i+12] += ry0*rx6 - iy0*ix6;
         A0[i+12+1] += ry0*ix6 + iy0*rx6;
         A1[i+12] += ry1*rx6 - iy1*ix6;
         A1[i+12+1] += ry1*ix6 + iy1*rx6;
         A2[i+12] += ry2*rx6 - iy2*ix6;
         A2[i+12+1] += ry2*ix6 + iy2*rx6;
         A3[i+12] += ry3*rx6 - iy3*ix6;
         A3[i+12+1] += ry3*ix6 + iy3*rx6;
         A0[i+14] += ry0*rx7 - iy0*ix7;
         A0[i+14+1] += ry0*ix7 + iy0*rx7;
         A1[i+14] += ry1*rx7 - iy1*ix7;
         A1[i+14+1] += ry1*ix7 + iy1*rx7;
         A2[i+14] += ry2*rx7 - iy2*ix7;
         A2[i+14+1] += ry2*ix7 + iy2*rx7;
         A3[i+14] += ry3*rx7 - iy3*ix7;
         A3[i+14+1] += ry3*ix7 + iy3*rx7;
         A0[i+16] += ry0*rx8 - iy0*ix8;
         prefA(A0);
         A0[i+16+1] += ry0*ix8 + iy0*rx8;
         A1[i+16] += ry1*rx8 - iy1*ix8;
         prefA(A1);
         A1[i+16+1] += ry1*ix8 + iy1*rx8;
         A2[i+16] += ry2*rx8 - iy2*ix8;
         prefA(A2);
         A2[i+16+1] += ry2*ix8 + iy2*rx8;
         A3[i+16] += ry3*rx8 - iy3*ix8;
         prefA(A3);
         A3[i+16+1] += ry3*ix8 + iy3*rx8;
         A0[i+18] += ry0*rx9 - iy0*ix9;
         A0[i+18+1] += ry0*ix9 + iy0*rx9;
         A1[i+18] += ry1*rx9 - iy1*ix9;
         A1[i+18+1] += ry1*ix9 + iy1*rx9;
         A2[i+18] += ry2*rx9 - iy2*ix9;
         A2[i+18+1] += ry2*ix9 + iy2*rx9;
         A3[i+18] += ry3*rx9 - iy3*ix9;
         A3[i+18+1] += ry3*ix9 + iy3*rx9;
         A0[i+20] += ry0*rx10 - iy0*ix10;
         A0[i+20+1] += ry0*ix10 + iy0*rx10;
         A1[i+20] += ry1*rx10 - iy1*ix10;
         A1[i+20+1] += ry1*ix10 + iy1*rx10;
         A2[i+20] += ry2*rx10 - iy2*ix10;
         A2[i+20+1] += ry2*ix10 + iy2*rx10;
         A3[i+20] += ry3*rx10 - iy3*ix10;
         A3[i+20+1] += ry3*ix10 + iy3*rx10;
         A0[i+22] += ry0*rx11 - iy0*ix11;
         A0[i+22+1] += ry0*ix11 + iy0*rx11;
         A1[i+22] += ry1*rx11 - iy1*ix11;
         A1[i+22+1] += ry1*ix11 + iy1*rx11;
         A2[i+22] += ry2*rx11 - iy2*ix11;
         A2[i+22+1] += ry2*ix11 + iy2*rx11;
         A3[i+22] += ry3*rx11 - iy3*ix11;
         A3[i+22+1] += ry3*ix11 + iy3*rx11;
         A0[i+24] += ry0*rx12 - iy0*ix12;
         prefA(A0);
         A0[i+24+1] += ry0*ix12 + iy0*rx12;
         A1[i+24] += ry1*rx12 - iy1*ix12;
         prefA(A1);
         A1[i+24+1] += ry1*ix12 + iy1*rx12;
         A2[i+24] += ry2*rx12 - iy2*ix12;
         prefA(A2);
         A2[i+24+1] += ry2*ix12 + iy2*rx12;
         A3[i+24] += ry3*rx12 - iy3*ix12;
         prefA(A3);
         A3[i+24+1] += ry3*ix12 + iy3*rx12;
         A0[i+26] += ry0*rx13 - iy0*ix13;
         A0[i+26+1] += ry0*ix13 + iy0*rx13;
         A1[i+26] += ry1*rx13 - iy1*ix13;
         A1[i+26+1] += ry1*ix13 + iy1*rx13;
         A2[i+26] += ry2*rx13 - iy2*ix13;
         A2[i+26+1] += ry2*ix13 + iy2*rx13;
         A3[i+26] += ry3*rx13 - iy3*ix13;
         A3[i+26+1] += ry3*ix13 + iy3*rx13;
         A0[i+28] += ry0*rx14 - iy0*ix14;
         A0[i+28+1] += ry0*ix14 + iy0*rx14;
         A1[i+28] += ry1*rx14 - iy1*ix14;
         A1[i+28+1] += ry1*ix14 + iy1*rx14;
         A2[i+28] += ry2*rx14 - iy2*ix14;
         A2[i+28+1] += ry2*ix14 + iy2*rx14;
         A3[i+28] += ry3*rx14 - iy3*ix14;
         A3[i+28+1] += ry3*ix14 + iy3*rx14;
         A0[i+30] += ry0*rx15 - iy0*ix15;
         A0[i+30+1] += ry0*ix15 + iy0*rx15;
         A1[i+30] += ry1*rx15 - iy1*ix15;
         A1[i+30+1] += ry1*ix15 + iy1*rx15;
         A2[i+30] += ry2*rx15 - iy2*ix15;
         A2[i+30+1] += ry2*ix15 + iy2*rx15;
         A3[i+30] += ry3*rx15 - iy3*ix15;
         A3[i+30+1] += ry3*ix15 + iy3*rx15;
         A0[i+32] += ry0*rx16 - iy0*ix16;
         prefA(A0);
         A0[i+32+1] += ry0*ix16 + iy0*rx16;
         A1[i+32] += ry1*rx16 - iy1*ix16;
         prefA(A1);
         A1[i+32+1] += ry1*ix16 + iy1*rx16;
         A2[i+32] += ry2*rx16 - iy2*ix16;
         prefA(A2);
         A2[i+32+1] += ry2*ix16 + iy2*rx16;
         A3[i+32] += ry3*rx16 - iy3*ix16;
         prefA(A3);
         A3[i+32+1] += ry3*ix16 + iy3*rx16;
         A0[i+34] += ry0*rx17 - iy0*ix17;
         A0[i+34+1] += ry0*ix17 + iy0*rx17;
         A1[i+34] += ry1*rx17 - iy1*ix17;
         A1[i+34+1] += ry1*ix17 + iy1*rx17;
         A2[i+34] += ry2*rx17 - iy2*ix17;
         A2[i+34+1] += ry2*ix17 + iy2*rx17;
         A3[i+34] += ry3*rx17 - iy3*ix17;
         A3[i+34+1] += ry3*ix17 + iy3*rx17;
         A0[i+36] += ry0*rx18 - iy0*ix18;
         A0[i+36+1] += ry0*ix18 + iy0*rx18;
         A1[i+36] += ry1*rx18 - iy1*ix18;
         A1[i+36+1] += ry1*ix18 + iy1*rx18;
         A2[i+36] += ry2*rx18 - iy2*ix18;
         A2[i+36+1] += ry2*ix18 + iy2*rx18;
         A3[i+36] += ry3*rx18 - iy3*ix18;
         A3[i+36+1] += ry3*ix18 + iy3*rx18;
         A0[i+38] += ry0*rx19 - iy0*ix19;
         A0[i+38+1] += ry0*ix19 + iy0*rx19;
         A1[i+38] += ry1*rx19 - iy1*ix19;
         A1[i+38+1] += ry1*ix19 + iy1*rx19;
         A2[i+38] += ry2*rx19 - iy2*ix19;
         A2[i+38+1] += ry2*ix19 + iy2*rx19;
         A3[i+38] += ry3*rx19 - iy3*ix19;
         A3[i+38+1] += ry3*ix19 + iy3*rx19;
         A0[i+40] += ry0*rx20 - iy0*ix20;
         prefA(A0);
         A0[i+40+1] += ry0*ix20 + iy0*rx20;
         A1[i+40] += ry1*rx20 - iy1*ix20;
         prefA(A1);
         A1[i+40+1] += ry1*ix20 + iy1*rx20;
         A2[i+40] += ry2*rx20 - iy2*ix20;
         prefA(A2);
         A2[i+40+1] += ry2*ix20 + iy2*rx20;
         A3[i+40] += ry3*rx20 - iy3*ix20;
         prefA(A3);
         A3[i+40+1] += ry3*ix20 + iy3*rx20;
         A0[i+42] += ry0*rx21 - iy0*ix21;
         A0[i+42+1] += ry0*ix21 + iy0*rx21;
         A1[i+42] += ry1*rx21 - iy1*ix21;
         A1[i+42+1] += ry1*ix21 + iy1*rx21;
         A2[i+42] += ry2*rx21 - iy2*ix21;
         A2[i+42+1] += ry2*ix21 + iy2*rx21;
         A3[i+42] += ry3*rx21 - iy3*ix21;
         A3[i+42+1] += ry3*ix21 + iy3*rx21;
         A0[i+44] += ry0*rx22 - iy0*ix22;
         A0[i+44+1] += ry0*ix22 + iy0*rx22;
         A1[i+44] += ry1*rx22 - iy1*ix22;
         A1[i+44+1] += ry1*ix22 + iy1*rx22;
         A2[i+44] += ry2*rx22 - iy2*ix22;
         A2[i+44+1] += ry2*ix22 + iy2*rx22;
         A3[i+44] += ry3*rx22 - iy3*ix22;
         A3[i+44+1] += ry3*ix22 + iy3*rx22;
         A0[i+46] += ry0*rx23 - iy0*ix23;
         A0[i+46+1] += ry0*ix23 + iy0*rx23;
         A1[i+46] += ry1*rx23 - iy1*ix23;
         A1[i+46+1] += ry1*ix23 + iy1*rx23;
         A2[i+46] += ry2*rx23 - iy2*ix23;
         A2[i+46+1] += ry2*ix23 + iy2*rx23;
         A3[i+46] += ry3*rx23 - iy3*ix23;
         A3[i+46+1] += ry3*ix23 + iy3*rx23;
         A0[i+48] += ry0*rx24 - iy0*ix24;
         prefA(A0);
         A0[i+48+1] += ry0*ix24 + iy0*rx24;
         A1[i+48] += ry1*rx24 - iy1*ix24;
         prefA(A1);
         A1[i+48+1] += ry1*ix24 + iy1*rx24;
         A2[i+48] += ry2*rx24 - iy2*ix24;
         prefA(A2);
         A2[i+48+1] += ry2*ix24 + iy2*rx24;
         A3[i+48] += ry3*rx24 - iy3*ix24;
         prefA(A3);
         A3[i+48+1] += ry3*ix24 + iy3*rx24;
         A0[i+50] += ry0*rx25 - iy0*ix25;
         A0[i+50+1] += ry0*ix25 + iy0*rx25;
         A1[i+50] += ry1*rx25 - iy1*ix25;
         A1[i+50+1] += ry1*ix25 + iy1*rx25;
         A2[i+50] += ry2*rx25 - iy2*ix25;
         A2[i+50+1] += ry2*ix25 + iy2*rx25;
         A3[i+50] += ry3*rx25 - iy3*ix25;
         A3[i+50+1] += ry3*ix25 + iy3*rx25;
         A0[i+52] += ry0*rx26 - iy0*ix26;
         A0[i+52+1] += ry0*ix26 + iy0*rx26;
         A1[i+52] += ry1*rx26 - iy1*ix26;
         A1[i+52+1] += ry1*ix26 + iy1*rx26;
         A2[i+52] += ry2*rx26 - iy2*ix26;
         A2[i+52+1] += ry2*ix26 + iy2*rx26;
         A3[i+52] += ry3*rx26 - iy3*ix26;
         A3[i+52+1] += ry3*ix26 + iy3*rx26;
         A0[i+54] += ry0*rx27 - iy0*ix27;
         A0[i+54+1] += ry0*ix27 + iy0*rx27;
         A1[i+54] += ry1*rx27 - iy1*ix27;
         A1[i+54+1] += ry1*ix27 + iy1*rx27;
         A2[i+54] += ry2*rx27 - iy2*ix27;
         A2[i+54+1] += ry2*ix27 + iy2*rx27;
         A3[i+54] += ry3*rx27 - iy3*ix27;
         A3[i+54+1] += ry3*ix27 + iy3*rx27;
         A0[i+56] += ry0*rx28 - iy0*ix28;
         prefA(A0);
         A0[i+56+1] += ry0*ix28 + iy0*rx28;
         A1[i+56] += ry1*rx28 - iy1*ix28;
         prefA(A1);
         A1[i+56+1] += ry1*ix28 + iy1*rx28;
         A2[i+56] += ry2*rx28 - iy2*ix28;
         prefA(A2);
         A2[i+56+1] += ry2*ix28 + iy2*rx28;
         A3[i+56] += ry3*rx28 - iy3*ix28;
         prefA(A3);
         A3[i+56+1] += ry3*ix28 + iy3*rx28;
         A0[i+58] += ry0*rx29 - iy0*ix29;
         A0[i+58+1] += ry0*ix29 + iy0*rx29;
         A1[i+58] += ry1*rx29 - iy1*ix29;
         A1[i+58+1] += ry1*ix29 + iy1*rx29;
         A2[i+58] += ry2*rx29 - iy2*ix29;
         A2[i+58+1] += ry2*ix29 + iy2*rx29;
         A3[i+58] += ry3*rx29 - iy3*ix29;
         A3[i+58+1] += ry3*ix29 + iy3*rx29;
         A0[i+60] += ry0*rx30 - iy0*ix30;
         A0[i+60+1] += ry0*ix30 + iy0*rx30;
         A1[i+60] += ry1*rx30 - iy1*ix30;
         A1[i+60+1] += ry1*ix30 + iy1*rx30;
         A2[i+60] += ry2*rx30 - iy2*ix30;
         A2[i+60+1] += ry2*ix30 + iy2*rx30;
         A3[i+60] += ry3*rx30 - iy3*ix30;
         A3[i+60+1] += ry3*ix30 + iy3*rx30;
         A0[i+62] += ry0*rx31 - iy0*ix31;
         A0[i+62+1] += ry0*ix31 + iy0*rx31;
         A1[i+62] += ry1*rx31 - iy1*ix31;
         A1[i+62+1] += ry1*ix31 + iy1*rx31;
         A2[i+62] += ry2*rx31 - iy2*ix31;
         A2[i+62+1] += ry2*ix31 + iy2*rx31;
         A3[i+62] += ry3*rx31 - iy3*ix31;
         A3[i+62+1] += ry3*ix31 + iy3*rx31;
      }
      for (i=M32; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         A0[i] += ry0 * rx0 - iy0 * ix0;
         A0[i+1] += ry0 * ix0 + iy0 * rx0;
         A1[i] += ry1 * rx0 - iy1 * ix0;
         A1[i+1] += ry1 * ix0 + iy1 * rx0;
         A2[i] += ry2 * rx0 - iy2 * ix0;
         A2[i+1] += ry2 * ix0 + iy2 * rx0;
         A3[i] += ry3 * rx0 - iy3 * ix0;
         A3[i+1] += ry3 * ix0 + iy3 * rx0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N4; j; j--, A += lda, Y += 2)
   {
      double *A0=A;
      const register double ry0=Y[0], iy0=Y[1];
      ATL_INT i;
      prefY(Y+8+8-1);
      for (i=0; i < M32; i += 64)
      {
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31], rx16=X[i+32],
                               ix16=X[i+33], rx17=X[i+34], ix17=X[i+35],
                               rx18=X[i+36], ix18=X[i+37], rx19=X[i+38],
                               ix19=X[i+39], rx20=X[i+40], ix20=X[i+41],
                               rx21=X[i+42], ix21=X[i+43], rx22=X[i+44],
                               ix22=X[i+45], rx23=X[i+46], ix23=X[i+47],
                               rx24=X[i+48], ix24=X[i+49], rx25=X[i+50],
                               ix25=X[i+51], rx26=X[i+52], ix26=X[i+53],
                               rx27=X[i+54], ix27=X[i+55], rx28=X[i+56],
                               ix28=X[i+57], rx29=X[i+58], ix29=X[i+59],
                               rx30=X[i+60], ix30=X[i+61], rx31=X[i+62],
                               ix31=X[i+63];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A0[i+2] += ry0*rx1 - iy0*ix1;
         A0[i+2+1] += ry0*ix1 + iy0*rx1;
         A0[i+4] += ry0*rx2 - iy0*ix2;
         A0[i+4+1] += ry0*ix2 + iy0*rx2;
         A0[i+6] += ry0*rx3 - iy0*ix3;
         A0[i+6+1] += ry0*ix3 + iy0*rx3;
         A0[i+8] += ry0*rx4 - iy0*ix4;
         prefA(A0);
         A0[i+8+1] += ry0*ix4 + iy0*rx4;
         A0[i+10] += ry0*rx5 - iy0*ix5;
         A0[i+10+1] += ry0*ix5 + iy0*rx5;
         A0[i+12] += ry0*rx6 - iy0*ix6;
         A0[i+12+1] += ry0*ix6 + iy0*rx6;
         A0[i+14] += ry0*rx7 - iy0*ix7;
         A0[i+14+1] += ry0*ix7 + iy0*rx7;
         A0[i+16] += ry0*rx8 - iy0*ix8;
         prefA(A0);
         A0[i+16+1] += ry0*ix8 + iy0*rx8;
         A0[i+18] += ry0*rx9 - iy0*ix9;
         A0[i+18+1] += ry0*ix9 + iy0*rx9;
         A0[i+20] += ry0*rx10 - iy0*ix10;
         A0[i+20+1] += ry0*ix10 + iy0*rx10;
         A0[i+22] += ry0*rx11 - iy0*ix11;
         A0[i+22+1] += ry0*ix11 + iy0*rx11;
         A0[i+24] += ry0*rx12 - iy0*ix12;
         prefA(A0);
         A0[i+24+1] += ry0*ix12 + iy0*rx12;
         A0[i+26] += ry0*rx13 - iy0*ix13;
         A0[i+26+1] += ry0*ix13 + iy0*rx13;
         A0[i+28] += ry0*rx14 - iy0*ix14;
         A0[i+28+1] += ry0*ix14 + iy0*rx14;
         A0[i+30] += ry0*rx15 - iy0*ix15;
         A0[i+30+1] += ry0*ix15 + iy0*rx15;
         A0[i+32] += ry0*rx16 - iy0*ix16;
         prefA(A0);
         A0[i+32+1] += ry0*ix16 + iy0*rx16;
         A0[i+34] += ry0*rx17 - iy0*ix17;
         A0[i+34+1] += ry0*ix17 + iy0*rx17;
         A0[i+36] += ry0*rx18 - iy0*ix18;
         A0[i+36+1] += ry0*ix18 + iy0*rx18;
         A0[i+38] += ry0*rx19 - iy0*ix19;
         A0[i+38+1] += ry0*ix19 + iy0*rx19;
         A0[i+40] += ry0*rx20 - iy0*ix20;
         prefA(A0);
         A0[i+40+1] += ry0*ix20 + iy0*rx20;
         A0[i+42] += ry0*rx21 - iy0*ix21;
         A0[i+42+1] += ry0*ix21 + iy0*rx21;
         A0[i+44] += ry0*rx22 - iy0*ix22;
         A0[i+44+1] += ry0*ix22 + iy0*rx22;
         A0[i+46] += ry0*rx23 - iy0*ix23;
         A0[i+46+1] += ry0*ix23 + iy0*rx23;
         A0[i+48] += ry0*rx24 - iy0*ix24;
         prefA(A0);
         A0[i+48+1] += ry0*ix24 + iy0*rx24;
         A0[i+50] += ry0*rx25 - iy0*ix25;
         A0[i+50+1] += ry0*ix25 + iy0*rx25;
         A0[i+52] += ry0*rx26 - iy0*ix26;
         A0[i+52+1] += ry0*ix26 + iy0*rx26;
         A0[i+54] += ry0*rx27 - iy0*ix27;
         A0[i+54+1] += ry0*ix27 + iy0*rx27;
         A0[i+56] += ry0*rx28 - iy0*ix28;
         prefA(A0);
         A0[i+56+1] += ry0*ix28 + iy0*rx28;
         A0[i+58] += ry0*rx29 - iy0*ix29;
         A0[i+58+1] += ry0*ix29 + iy0*rx29;
         A0[i+60] += ry0*rx30 - iy0*ix30;
         A0[i+60+1] += ry0*ix30 + iy0*rx30;
         A0[i+62] += ry0*rx31 - iy0*ix31;
         A0[i+62+1] += ry0*ix31 + iy0*rx31;
      }
      for (i=M32; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         A0[i] += ry0 * rx0 - iy0 * ix0;
         A0[i+1] += ry0 * ix0 + iy0 * rx0;
      }
   }
}
